import hashlib
import numpy as np
from pathlib import Path


def list_to_coll(list_input: list, out_file: str) -> None:
    """Turns a list into a coll."""
    f = open(out_file, "w+")
    counter = 0
    for item in list_input:
        f.write(f"{counter}, {item};")
        counter += 1
    f.close()


def bytes_to_mb(val: int) -> float:
    """convert bytes to mb"""
    return val * 0.000001


def create_hash(*items) -> str:
    """Create a hash from a list of items"""
    m = hashlib.blake2b(digest_size=20)
    for item in items:
        m.update(str(item).encode("utf-8"))
    return m.hexdigest()

ignored_keys = (  # keys to ignore from superclass
    "process",
    "dump_path",
    "model_dump",
    "input",
    "output",
    "input_type",
    "dump_type",
    "cache",
    "cache_possible",
    "corpus_items",
    "buffer",
    "pre", "post",
    "scripting", "chain", "parent", "parent_string"
)



from ftis.common.analyser import FTISAnalyser
from ftis.common.io import write_json, read_json
from ftis.common.proc import staticproc
from umap import UMAP as umapdr
from joblib import dump as jdump
import numpy as np


class UMAP(FTISAnalyser):
    """Dimension reduction with UMAP algorithm"""

    def __init__(self, mindist=0.01, neighbours=7, components=2, cache=False):
        super().__init__(cache=cache)
        self.mindist = mindist
        self.neighbours = neighbours
        self.components = components
        self.output = {}
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        jdump(self.model, self.model_dump)
        write_json(self.dump_path, self.output)

    def analyse(self):
        data = [v for v in self.input.values()]
        keys = [k for k in self.input.keys()]

        data = np.array(data)

        self.model = umapdr(
            n_components=self.components, n_neighbors=self.neighbours, min_dist=self.mindist, random_state=42
        )
        self.model.fit(data)
        transformed_data = self.model.transform(data)
        self.output = {k: v.tolist() for k, v in zip(keys, transformed_data)}

    def run(self):
        staticproc(self.name, self.analyse)



import datetime
import logging
from pathlib import Path
from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table
from rich.text import Text
from rich import box
from ftis.common.io import write_json, read_json
from ftis.common.utils import ignored_keys, create_hash
from ftis.corpus import Corpus
from shutil import rmtree

class World:
    def __init__(self, source=None, sink=None, quiet=False, clear=False):
        self.sink = Path(sink).expanduser().resolve()
        self.node_depth = 0
        # Input corpora objects
        self.corpora = []
        # Metadata
        self.metadata = {"analyser" : {}}
        self.prev_meta = None
        self.clear = clear
        # Console
        self.console = Console()
        self.quiet = quiet
        # Logging
        self.logger = logging.getLogger(__name__)

    def setup(self) -> None:
        self.metadata["time"] = datetime.datetime.now().strftime("%H:%M:%S | %B %d, %Y")
        self.sink.mkdir(exist_ok=True, parents=True)
        
        # Microcache
        self.cache = self.sink / ".cache"
        self.cache.mkdir(exist_ok=True)

        # Setup logging and meta path
        self.metapath = self.sink / "metadata.json"
        logfile_path = self.sink / "logfile.log"

        # Load previous meta data if around
        if self.metapath.exists() and self.metapath.is_file():
            self.prev_meta = read_json(self.metapath)

        # Init loggin
        self.logger.setLevel(logging.DEBUG)

        if logfile_path.exists():
            logfile_path.unlink()

        logfile_handler = logging.FileHandler(logfile_path)
        formatter = logging.Formatter("%(asctime)s : %(levelname)s : %(name)s : %(message)s")
        logfile_handler.setFormatter(formatter)
        self.logger.addHandler(logfile_handler)
        self.logger.debug("Logging initialised")

    def build_connections(self, node):
        node.process = self # set the process to the world
        if not isinstance(node, Corpus) and not isinstance(node, World):
            node.create_identity()
            node.set_dump()
        for suborder, child in enumerate(node.chain):
            # Set the parent of the children to the node passed in
            child.parent = node
            # Establish the order
            if isinstance(child.parent, Corpus):
                child.order = 1
            else:
                child.order = child.parent.order + 1
                child.suborder = suborder

            self.build_connections(child)
        self.metadata["analyser"][node.__class__.__name__] = {
            k: str(v) 
            for k, v in vars(node).items() 
            if k not in ignored_keys
        }
        
    def build(self, *corpora):
        self.corpora = corpora
        self.setup()
        # This is a two stage process hence two loops.
        # 1: Build a graph (including depth) of connections
        for c in corpora:
            self.build_connections(c)

    def run(self):
        if not self.quiet:
            version = "# **** FTIS v2.0.0a ****"
            self.console.print(Markdown(version))
            # Construct information table
            corpora_paths = Table(title="Corpora", box=box.HORIZONTALS, show_lines=True)
            corpora_paths.add_column("Path")
            corpora_paths.add_column("Items", style="cyan")
            for c in self.corpora:
                corpora_paths.add_row(str(c.path), str(len(c.items)))
        
            print("\n")
            self.console.print(corpora_paths)
            sink_text = Text(f" Sink: {self.sink}")
            print("\n")
            self.console.print(sink_text)
            self.console.print(Markdown("---"))
            print("\n")

        for c in self.corpora:
            c.walk_chain()

        self.teardown()

    def teardown(self):
        write_json(self.metapath, self.metadata)
        if self.clear:
            self.clear_cache()

    def fprint(self, text):
        self.console.print(text, style="yellow underline")

    def clear_cache(self) -> None:
        rmtree(str(self.cache))



import numpy as np
from flucoma.fluid import loudness, stats
from flucoma.utils import get_buffer
from ftis.common.proc import staticproc
from ftis.common.utils import create_hash
from ftis.common.io import get_duration, write_json, read_json
from ftis.common.analyser import FTISAnalyser
from pathlib import Path


class Duration(FTISAnalyser):
    def __init__(self, min_dur=0, max_dur=36000, cache=False):  # 10 hours of audio
        super().__init__(cache=cache)
        self.output = []
        self.min_dur: int = min_dur
        self.max_dur: int = max_dur
        self.dump_type = ".json"

    def load_cache(self):
        d = read_json(self.dump_path)
        self.output = [Path(x) for x in d["corpus_items"]]

    def dump(self):
        d = {"corpus_items": [str(x) for x in self.output]}
        write_json(self.dump_path, d)

    def filter_duration(self, x):
        dur = get_duration(x)
        return dur < self.max_dur and dur > self.min_dur

    def filter_items(self):
        self.output = [x for x in self.input if self.filter_duration(x)]

    def run(self):
        staticproc(self.name, self.filter_items)


class Loudness(FTISAnalyser):
    """A way to filter corpus items after they have been loaded by the corpus loader"""

    def __init__(self, min_loudness=0, max_loudness=100, cache=False):
        super().__init__(cache=cache)
        self.min_loudness: float = min_loudness
        self.max_loudness: float = max_loudness
        self.dump_type = ".json"

    def load_cache(self):
        d = read_json(self.dump_path)
        self.output = [Path(x) for x in d["corpus_items"]]

    def dump(self):
        d = {"corpus_items": [str(x) for x in self.output]}
        write_json(self.dump_path, d)

    def analyse_items(self):
        # TODO import this function from a global place that can be used in CORPUS too
        median_loudness = {}
        for x in self.input:
            hsh = create_hash(x, self.min_loudness, self.max_loudness)
            cache = self.process.cache / f"{hsh}.npy"
            if not cache.exists():
                med_loudness = get_buffer(stats(loudness(x, hopsize=4410, windowsize=17640)), "numpy")
                np.save(cache, med_loudness)
            else:
                med_loudness = np.load(cache, allow_pickle=True)

            median_loudness[str(x)] = med_loudness[0][5]
        vals = np.array([x for x in median_loudness.values()])
        min_perc = np.percentile(vals, self.min_loudness)
        max_perc = np.percentile(vals, self.max_loudness)
        self.output = [k for k, v in median_loudness.items() if v <= max_perc and v >= min_perc]

    def run(self):
        staticproc(self.name, self.analyse_items)



from ftis.common.analyser import FTISAnalyser
from ftis.common.io import write_json, read_json
from ftis.common.proc import singleproc
from multiprocessing import Manager
from scipy.stats import describe
from math import sqrt
import numpy as np


class Stats(FTISAnalyser):
    """Get various statistics and derivatives of those"""

    def __init__(
        self,
        numderivs=0,
        flatten=True,
        spec=["mean", "stddev", "skewness", "kurtosis", "min", "median", "max"],
        cache=False,
    ):

        super().__init__(cache=cache)
        self.numderivs = numderivs
        self.flatten = flatten
        self.spec = spec
        self.dump_type = ".json"

    def dump(self):
        write_json(self.dump_path, self.output)

    def load_cache(self):
        self.output = read_json(self.dump_path)

    @staticmethod
    def calc_stats(data, spec):
        description = describe(data)
        output = []
        if "mean" in spec:
            output.append(description.mean)
        if "stddev" in spec:
            output.append(sqrt(description.variance))
        if "skewness" in spec:
            output.append(description.skewness)
        if "kurtosis" in spec:
            output.append(description.kurtosis)
        if "minimum" in spec:
            output.append(description.minmax[0])
        if "median" in spec:
            output.append(np.median(data))
        if "maximum" in spec:
            output.append(description.minmax[1])
        return output

    def get_stats(self, base_data, num_derivs: int) -> list:
        """Given stats on n number derivatives from initial data"""
        container = []
        if num_derivs > 0:
            for i in range(num_derivs):
                deriv = np.diff(base_data, i + 1)
                container.append(self.calc_stats(deriv, self.spec))
        else:
            container = self.calc_stats(base_data, self.spec)
        return container

    def analyse(self, workable):
        # TODO: any dimensionality input
        element_container = []
        values = np.array(self.input[workable])
        if len(values.shape) < 2:  # single row we run the stats on that
            element_container.append(self.get_stats(values, self.numderivs))
        else:
            for row in values:  # for mfcc band in mfcc
                element_container.append(self.get_stats(row, self.numderivs))

        if self.flatten:
            element_container = np.array(element_container)
            element_container = element_container.flatten()
            element_container = element_container.tolist()
        self.buffer[workable] = element_container

    def run(self):
        self.buffer = Manager().dict()
        workables = [x for x in self.input.keys()]
        singleproc(self.name, self.analyse, workables)
        self.output = dict(self.buffer)



def samps2ms(samples: float, sr: int) -> float:
    """samples to milliseconds given a sampling rate"""
    return (samples / sr) * 1000.0


def ms2samps(ms: int, sr: int) -> int:
    """milliseconds to samples given a sample rate"""
    return (ms * 0.001) * sr


def secs2ms(seconds: int) -> int:
    """seconds to milliseconds"""
    return seconds * 1000



from ftis.common.exceptions import OutputNotFound
from ftis.common.types import ftypes
from ftis.common.io import read_json, write_json
from ftis.common.utils import ignored_keys, create_hash
from collections.abc import Callable
from collections import OrderedDict
from pathlib import Path


class FTISAnalyser:
    """Every analyser inherits from this class"""
    def __init__(self, cache=False, pre=None, post=None):
        self.process = None  # pass the parent process in
        self.input = None  # This can be anything
        self.output = None
        self.input_type: str = ""
        self.dump_type: str = ""
        self.dump_path: Path = None
        self.model_dump: Path = None  #
        self.name = self.__class__.__name__
        self.order: int = -1
        self.cache: bool = cache
        self.cache_possible: bool = False
        self.pre: Callable = pre
        self.post: Callable = post
        # Overloading Stuff
        self.scripting = True
        self.suborder = 0
        self.parent = None
        self.chain = OrderedDict()
        self.parent_string = self.__class__.__name__

    def __str__(self):
        return f"{self.__class__.__name__}"

    def __rshift__(self, right):
        # right.order = self.order + 1
        self.scripting = True
        self.chain[right] = None
        return right
    
    def traverse_parent_parameters(self):
        self.parent_parameters[self.parent.__class__.__name__] = ({
            k: v 
            for k, v in vars(self).items() 
            if k not in ignored_keys
            })
        if hasattr(self.parent, 'parent'): # if the parent has a parent
            self.parent.traverse_parent_parameters()

    def create_identity(self) -> None:
        if not self.scripting:
            self.identity = {k: v for k, v in vars(self).items() if k not in ignored_keys}
            previous_inputs = {}
            for obj in self.process.chain:
                previous_inputs[str(obj.name)] = {k: v for k, v in vars(obj).items() if k not in ignored_keys}

            self.identity_hash = create_hash(previous_inputs)
            self.identity["identity_hash"] = self.identity_hash
        else:
            self.identity = {
                k: v 
                for k, v in vars(self).items() 
                if k not in ignored_keys
            }
            self.parent_parameters = {}
            self.traverse_parent_parameters()
            self.identity["hash"] = create_hash(self.parent_parameters)

    def log(self, log_text: str) -> None:
        try:
            self.process.logger.debug(f"{self.name}: {log_text}")
        except AttributeError:
            pass

    def _get_parents(self) -> None:
        self.parent_string = (
            f"{self.parent.__class__.__name__}.{self.parent_string}"
        )

    def set_dump(self) -> None:
        self._get_parents()
        if self.scripting:
            self.dump_path  = (
                self.process.sink / 
                f"{self.order}.{self.suborder}-{self.parent_string}{self.dump_type}"
            )
            self.model_dump = (
                self.process.sink / 
                f"{self.order}.{self.suborder}-{self.parent_string}.joblib"
            )
        else:
            pass


    def folder_integrity(self) -> bool:
        # TODO: implement folder integrity checking for analysers like Explode/Collapse
        # TODO: Implement a method for knowing about folder-y outputs before they're made (workables!)
        return True

    def compare_meta(self) -> bool:
        # TODO You could use a hashing function here to determine the similarity of the metadata
        # TODO You should use a hashing function because adding things to the front of the chain makes it not equal between runs
        self.process.metadata = self.process.metadata
        self.process.prev_meta = self.process.prev_meta
        ident = f"{self.order}_{self.name}"
        try:
            new_params = self.process.metadata["analyser"][ident]
        except KeyError:
            new_params = False

        try:
            old_params = self.process.prev_meta["analyser"][ident]
        except KeyError:
            old_params = False

        try:
            success = self.process.prev_meta["success"][ident]
        except KeyError:
            success = False

        return old_params == new_params and success

    def cache_exists(self) -> bool:
        if self.dump_path.exists():
            if self.dump_type == ftypes.folder:
                return self.folder_integrity()
            else:
                return True
        else:
            return False

    def update_success(self, status: bool) -> None:
        try:
            existing_metadata = read_json(self.process.metapath)
        except FileNotFoundError:
            existing_metadata = {}

        try:
            success = existing_metadata["success"]  # extract the progress dict
        except KeyError:
            success = {}  # progress doesnt exist yet

        success[f"{self.order}_{self.name}"] = status  # update the status of this analyser
        # join any existing data into the metadata
        self.process.metadata["success"] = success  # modify the original
        write_json(self.process.metapath, self.process.metadata)

    def do(self) -> None:
        self.log("Initiating")

        # Determine whether we caching is possible
        if self.cache and self.cache_exists() and self.compare_meta() and self.process.metapath.exists():
            self.cache_possible = True

        # Set the status to failure and only update to success if it all ends correctly
        self.update_success(False)
        if self.cache_possible:
            self.load_cache()
            self.process.fprint(f"{self.name} was cached")
        else:
            if self.pre:
                self.pre(self)
            self.run()
            if self.post:
                self.post(self)
            self.dump()

        if self.output != None:  # TODO comprehensive output checking
            self.log("Ran Successfully")
            self.update_success(True)
        else:
            self.log("Output was invalid")
            raise OutputNotFound(self.name)

    def walk_chain(self) -> None:
        self.log("Initialising")
        # Determine whether we caching is possible
        if self.cache and self.cache_exists() and self.compare_meta() and self.process.metapath.exists():
            self.cache_possible = True
        
        self.update_success(False)
        if self.cache_possible:
            self.load_cache()
            self.process.fprint(f"{self.name} was cached")
        else:
            if self.pre: # preprocess
                self.pre(self)
            self.run()
            if self.post: # postprocess
                self.post(self)
                self.dump()

        if self.output != None:  # TODO comprehensive output checking
            self.log("Ran Successfully")
            self.update_success(True)
        else:
            self.log("Output was invalid")
            raise OutputNotFound(self.name)

        self.dump()
        # Pass output to the input of all of connected things
        for forward_connection in self.chain:
            forward_connection.input = self.output
            forward_connection.walk_chain()

    def load_cache(self) -> None:
        """Implemented in the analyser"""

    def dump(self) -> None:
        """Defined in the analyser that inherits this class"""

    def run(self) -> None:
        """Method for running the processing chain from input to output"""



from concurrent.futures import ThreadPoolExecutor, as_completed
from rich.progress import Progress, BarColumn
from ftis.common.exceptions import EmptyWorkables


def multiproc(name: str, process, workables):
    """This function wraps up a multithreaded worker and progress bar"""

    if len(workables) == 0:
        raise EmptyWorkables

    with Progress() as progress:
        task = progress.add_task(name, total=len(workables))
        with ThreadPoolExecutor() as pool:
            futures = [pool.submit(process, work) for work in workables]
            for result in as_completed(futures):
                progress.update(task, advance=1)


def singleproc(name: str, process, workables):
    """This function wraps up a multithreaded worker and progress bar"""
    if len(workables) == 0:
        raise EmptyWorkables

    with Progress() as progress:
        task = progress.add_task(name, total=len(workables))
        for x in workables:
            process(x)
            progress.update(task, advance=1)


def staticproc(name: str, process):
    """For processes where progress can not be determined"""
    with Progress(f"[magenta]{name}...", BarColumn()) as progress:
        task = progress.add_task(name, start=False)
        process()



from ftis.common.analyser import FTISAnalyser
from ftis.common.io import write_json, read_json
from ftis.common.proc import multiproc, singleproc
from ftis.common.utils import create_hash
from multiprocessing import Manager
from flucoma.utils import get_buffer
from flucoma import fluid


class FluidOnsetslice(FTISAnalyser):
    def __init__(
        self,
        fftsettings=[1024, 512, 1024],
        filtersize=5,
        framedelta=0,
        metric=0,
        minslicelength=2,
        threshold=0.5,
        cache=False,
    ):
        super().__init__(cache=cache)
        self.fftsettings = fftsettings
        self.filtersize = filtersize
        self.framedelta = framedelta
        self.metric = metric
        self.minslicelength = minslicelength
        self.threshold = threshold
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def analyse(self, workable):
        hsh = create_hash(workable, self.identity)
        cache = self.process.cache / f"{hsh}.wav"
        if not cache.exists():
            slice_output = get_buffer(
                fluid.onsetslice(
                    workable,
                    indices=cache,
                    fftsettings=self.fftsettings,
                    filtersize=self.filtersize,
                    framedelta=self.framedelta,
                    metric=self.metric,
                    minslicelength=self.minslicelength,
                    threshold=self.threshold,
                ),
                "numpy",
            )
        else:
            slice_output = get_buffer(cache, "numpy")

        self.buffer[str(workable)] = slice_output.tolist()

    def run(self):
        self.buffer = Manager().dict()
        singleproc(self.name, self.analyse, self.input)
        self.output = dict(self.buffer)


class FluidNoveltyslice(FTISAnalyser):
    def __init__(
        self,
        feature=0,
        fftsettings=[1024, 512, 1024],
        filtersize=1,
        minslicelength=2,
        threshold=0.5,
        cache=False,
    ):
        super().__init__(cache=cache)
        self.feature = feature
        self.fftsettings = fftsettings
        self.filtersize = filtersize
        self.minslicelength = minslicelength
        self.threshold = threshold
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def analyse(self, workable):
        noveltyslice = fluid.noveltyslice(
            workable,
            feature=self.feature,
            fftsettings=self.fftsettings,
            filtersize=self.filtersize,
            minslicelength=self.minslicelength,
            threshold=self.threshold,
        )
        self.buffer[str(workable)] = get_buffer(noveltyslice)

    def run(self):
        self.buffer = Manager().dict()
        multiproc(self.name, self.analyse, self.input)
        self.output = dict(self.buffer)



import soundfile as sf
import audioread
from typing import Union, Tuple, List
from pathlib import Path
import json


def write_json(json_file_path: str, in_dict: dict) -> None:
    """Takes a dictionary and writes it to JSON file"""
    with open(json_file_path, "w+") as fp:
        json.dump(in_dict, fp, indent=4)


def read_json(json_file_path: str) -> dict:
    """Takes a JSON file and returns a dictionary"""
    with open(json_file_path, "r") as fp:
        data = json.load(fp)
        return data


def peek(audio_file_path: Union[str, Path], output: str = "np"):
    """
    Returns a tuple of audio data and its sampling rate
    The audio data can be a numpy array or list
    """
    data, sr = sf.read(audio_file_path, dtype="float32")
    data = data.transpose()
    if output == "list":
        return data.tolist(), sr
    if output == "np":
        return data, sr


def get_duration(path: Union[str, Path]) -> float:
    data, sr = peek(path)
    return len(data) / sr


def get_sr(path: Union[str, Path]) -> int:
    try:
        with sf.SoundFile(path) as f:
            return int(f.samplerate)
    except RuntimeError:
        with audioread.audio_open(path) as f:
            return int(f.samplerate)



from ftis.common.analyser import FTISAnalyser
from ftis.common.proc import staticproc, multiproc, singleproc
from ftis.common.io import write_json, read_json
from ftis.common.utils import create_hash
from flucoma.utils import get_buffer
from flucoma import fluid
from multiprocessing import Manager
import numpy as np
import librosa

class Flux(FTISAnalyser):
    def __init__(self, windowsize=1024, hopsize=512, cache=False):
        super().__init__(cache=cache)
        self.windowsize = windowsize
        self.hopsize = hopsize
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def flux(self, workable):
        hsh = create_hash(workable, self.identity)
        cache = self.process.cache / f"{hsh}.npy"

        if not cache.exists():   
            y, _= librosa.load(workable)
            fft = librosa.stft(y, win_length=self.windowsize, hop_length=self.hopsize)
            flux = np.sum(np.abs(np.diff(np.abs(fft))), axis=0)
            np.save(cache, flux)
        else:
            flux = np.load(cache)
        self.buffer[str(workable)] = flux.tolist()
    
    def run(self):
        self.buffer = Manager().dict()
        multiproc(self.name, self.flux, self.input)
        self.output = dict(self.buffer)

class Chroma(FTISAnalyser):
    def __init__(self, 
    numchroma=12,
    numoctaves=7,
    bins_per_octave=12,
    hop_length=512,
    fmin=None,
    cache=False):
        super().__init__(cache=cache)
        self.numchroma = numchroma
        self.hopsize = hop_length
        self.fmin = fmin
        self.numoctaves = numoctaves
        self.bins_per_octave = bins_per_octave
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def chroma(self, workable):
        hsh = create_hash(workable, self.identity)
        cache = self.process.cache / f"{hsh}.npy"

        if not cache.exists():
            y, sr = librosa.load(workable)
            chroma = librosa.feature.chroma_cqt(y=y, sr=sr)
            np.save(cache, chroma)
        else:
            chroma = np.load(cache)
        self.buffer[str(workable)] = chroma.tolist()

    def run(self):
        self.buffer = Manager().dict()
        multiproc(self.name, self.chroma, self.input)
        self.output = dict(self.buffer)

class FluidLoudness(FTISAnalyser):
    def __init__(self, windowsize=17640, hopsize=4410, kweighting=1, truepeak=1, 
        cache=False,
        pre=None,
        post=None):
        super().__init__(cache=cache, pre=pre, post=post)
        self.windowsize = windowsize
        self.hopsize = hopsize
        self.kweighting = kweighting
        self.truepeak = truepeak
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def analyse(self, workable):
        hsh = create_hash(workable, self.identity)
        cache = self.process.cache / f"{hsh}.npy"

        if not cache.exists():
            loudness = get_buffer(
                fluid.loudness(
                    workable,
                    windowsize=self.windowsize,
                    hopsize=self.hopsize,
                    kweighting=self.kweighting,
                    truepeak=self.truepeak,
                ),
                "numpy",
            )
            np.save(cache, loudness)
        else:
            loudness = np.load(cache, allow_pickle=True)
        self.buffer[str(workable)] = loudness.tolist()

    def run(self):
        self.buffer = Manager().dict()
        workables = self.input
        multiproc(self.name, self.analyse, workables)
        self.output = dict(self.buffer)


class FluidPitch(FTISAnalyser):
    def __init__(self, 
        algorithm=2,
        minfreq=20,
        maxfreq=10000.0,
        unit=0,
        fftsettings=[1024, -1, -1],
        cache=False,
        pre=None,
        post=None):
        super().__init__(cache=cache, pre=pre, post=post)
        self.algorithm=algorithm
        self.minfreq=minfreq
        self.maxfreq=maxfreq
        self.unit=unit
        self.fftsettings=fftsettings
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def analyse(self, workable):
        hsh = create_hash(workable, self.identity)
        cache = self.process.cache / f"{hsh}.npy"

        if not cache.exists():
            pitch = get_buffer(
                fluid.pitch(
                    workable,
                    algorithm=self.algorithm,
                    minfreq=self.minfreq,
                    maxfreq=self.maxfreq,
                    unit=self.unit,
                    fftsettings=self.fftsettings
                ),
                "numpy",
            )
            np.save(cache, pitch)
        else:
            pitch = np.load(cache, allow_pickle=True)
        self.buffer[str(workable)] = pitch.tolist()

    def run(self):
        self.buffer = Manager().dict()
        workables = self.input
        singleproc(self.name, self.analyse, workables)
        self.output = dict(self.buffer)


class FluidMFCC(FTISAnalyser):
    def __init__(
        self,
        fftsettings=[1024, 512, 1024],
        numbands=40,
        numcoeffs=13,
        minfreq=80,
        maxfreq=20000,
        discard=False,
        cache=False,
    ):
        super().__init__(cache=cache)
        self.fftsettings = fftsettings
        self.numbands = numbands
        self.numcoeffs = numcoeffs
        self.minfreq = minfreq
        self.maxfreq = maxfreq
        self.discard = discard
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def analyse(self, workable):
        hsh = create_hash(workable, self.identity)
        cache = self.process.cache / f"{hsh}.npy"
        if cache.exists():
            f = np.load(cache, allow_pickle=True)
        else:
            f = get_buffer(
                fluid.mfcc(
                    workable,
                    fftsettings=self.fftsettings,
                    numbands=self.numbands,
                    numcoeffs=self.numcoeffs,
                    minfreq=self.minfreq,
                    maxfreq=self.maxfreq,
                ),
                "numpy",
            )
            np.save(cache, f)
        if self.discard:
            self.buffer[str(workable)] = f.tolist()[1:]
        else:
            self.buffer[str(workable)] = f.tolist()

    def run(self):
        self.buffer = Manager().dict()
        multiproc(self.name, self.analyse, self.input)
        self.output = dict(self.buffer)


class LibroMFCC(FTISAnalyser):
    def __init__(
        self,
        numbands=40,
        numcoeffs=20,
        minfreq=80,
        maxfreq=20000,
        window=2048,
        hop=512,
        dct=2,
        cache=False,
    ):
        super().__init__(cache=cache)
        self.numbands = numbands
        self.numcoeffs = numcoeffs
        self.minfreq = minfreq
        self.maxfreq = maxfreq
        self.window = window
        self.hop = hop
        self.dct = dct
        self.discard = discard
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def analyse(self, workable):
        hsh = create_hash(workable, self.identity)
        cache = self.process.cache / f"{hsh}.npy"
        if cache.exists():
            feature = np.load(cache, allow_pickle=True)
            print("loaded cache")
        else:
            y, sr = librosa.load(workable, sr=None, mono=True)
            feature = librosa.feature.mfcc(
                y=y,
                sr=sr,
                n_mfcc=self.numcoeffs,
                dct_type=self.dct,
                n_mels=self.numbands,
                fmax=self.maxfreq,
                fmin=self.minfreq,
                hop_length=self.hop,
                n_fft=self.window,
            )
            np.save(cache, feature)

    def run(self):
        self.buffer = Manager().dict()
        singleproc(self.name, self.analyse, self.input)
        self.output = dict(self.buffer)


class LibroCQT(FTISAnalyser):
    def __init__(
        self,
        hop_length=512,
        minfreq=110,
        n_bins=84,
        bins_per_octave=12,
        tuning=0.0,
        filter_scale=1,
        norm=1,
        sparsity=0.01,
        window="hann",
        scale=True,
        pad_mode="reflect",
        cache=False,
    ):
        super().__init__(cache=cache)
        self.hop_length = hop_length
        self.minfreq = minfreq
        self.n_bins = n_bins
        self.bins_per_octave = bins_per_octave
        self.tuning = tuning
        self.filter_scale = filter_scale
        self.norm = norm
        self.sparsity = sparsity
        self.window = window
        self.scale = scale
        self.pad_mode = pad_mode
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def analyse(self, workable):
        hsh = create_hash(workable, self.identity)
        cache = self.process.cache / f"{hsh}.npy"
        if cache.exists():
            cqt = np.load(cache, allow_pickle=True)
        else:
            y, sr = librosa.load(workable, sr=None, mono=True)
            cqt = librosa.cqt(y, sr,
                fmin=self.minfreq,
                n_bins=self.n_bins,
                bins_per_octave=self.bins_per_octave,
                tuning=self.tuning,
                filter_scale=self.filter_scale,
                norm=self.norm,
                sparsity=self.sparsity,
                window=self.window,
                scale=self.scale,
                pad_mode=self.pad_mode,
            )
            np.save(cache, cqt)

        self.buffer[str(workable)] = np.abs(cqt).tolist()

    def run(self):
        self.buffer = Manager().dict()
        singleproc(self.name, self.analyse, self.input)
        self.output = dict(self.buffer)



import os
from shutil import copytree, rmtree, copyfile
from pathlib import Path
from ftis.common.analyser import FTISAnalyser
from ftis.common.io import write_json
from subprocess import call


class Visualiser(FTISAnalyser):
    def __init__(self, cache=False):
        super().__init__(cache=cache)
        self.data: dict = {"data": []}

    def check_dimensions(self):
        self.vals = [v for v in self.input.values()]
        self.keys = [k for k in self.input.keys()]

    # def copy_audio(self):
    #     for k in self.keys:
    #         copyfile()

    def fmt(self):
        """Format JSON for threejs"""
        for k, v in zip(self.keys, self.vals):
            d = {"name": k, "x": v[0], "y": v[1], "z": v[2]}
            self.data["data"].append(d)

    def run(self):
        self.output = self.process.sink / f"{self.order}_{self.__class__.__name__}"
        self.output.mkdir(exist_ok=True)
        self.check_dimensions()
        # self.copy_audio()
        self.fmt()
        script = Path(__file__).resolve()
        # copy data and assets
        web_assets = script.parent / "web_assets"
        dest = self.output / "web_assets"
        if dest.exists():
            rmtree(dest)
        copytree(web_assets, self.output / "web_assets")
        write_json(self.output / "web_assets" / "plot.json", self.data)

        # cmd = [
        #     "cd",
        #     str(self.output / "web_assets" / "html"),
        #     "&&"
        #     "python",
        #     "-m",
        #     "http.server"
        # ]
        # f = call(cmd)
        # print(f)



from ftis.common.analyser import FTISAnalyser
from ftis.common.io import write_json, read_json
from ftis.common.proc import staticproc
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from joblib import dump as jdump
import numpy as np


class Normalise(FTISAnalyser):
    def __init__(self, minimum=0, maximum=1, cache=False):
        super().__init__(cache=cache)
        self.min = minimum
        self.max = maximum
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def analyse(self):
        scaled_data = MinMaxScaler((self.min, self.max)).fit_transform(self.features)

        self.output = {}
        for k, v in zip(self.keys, scaled_data):
            self.output[k] = v.tolist()

    def run(self):
        self.keys = [x for x in self.input.keys()]
        self.features = [x for x in self.input.values()]
        staticproc(self.name, self.analyse)


class Standardise(FTISAnalyser):
    def __init__(self, cache=False):
        super().__init__(cache=cache)
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        jdump(self.model, self.model_dump)
        write_json(self.dump_path, self.output)

    def analyse(self):
        self.model = StandardScaler()
        self.model.fit(self.features)
        scaled_data = self.model.transform(self.features)
        self.output = {k: list(v) for k, v in zip(self.keys, scaled_data)}

    def run(self):
        self.keys = [k for k in self.input.keys()]
        self.features = [v for v in self.input.values()]
        staticproc(self.name, self.analyse)



from ftis.common.analyser import FTISAnalyser
from ftis.common.proc import singleproc
from ftis.common.io import write_json

class Dummy(FTISAnalyser):
    def __init__(self, cache=False):
        super().__init__(cache=cache)
        self.a = 1
        self.b = "foo"

    def dump(self):
        write_json(self.dump_path, self.output)

    def go(self, workable):
        self.output = {
            "1" : self.a,
            "2" : self.b
        }

    def run(self):
        singleproc(self.name, self.go, self.input)



class InvalidYamlError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class OutputNotFound(Exception):
    def __init__(self, analyser: str):
        super().__init__(f"Output not found or errored for {analyser}")


class InvalidSource(Exception):
    def __init__(self, source: str):
        super().__init__(f"{source} does not exist or is invalid")


class NoCorpusSource(Exception):
    def __init__(self, source: str):
        super().__init__(f"{source} does not exist or is invalid")


class BadCorpusSource(Exception):
    def __init__(self):
        super().__init__("The corpus source is not PURE")  # TODO make this more descriptive


class AnalyserNotFound(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class AnalyserParameterInvalid(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class NotYetImplemented(Exception):
    def __init__(self):
        super().__init__("This function is not implemented yet")


class ChainIOError(Exception):
    def __init__(self, analyser1, analyser2):
        super().__init__(
            f"""
        Incompatible types
        {analyser1.name} input: {analyser1.input_type}
        {analyser2.name} output: {analyser2.output_type}
        """
        )


class SourceIOError(Exception):
    def __init__(self):
        super().__init__(f"The source and first analyser type are incompatible")


class EmptyWorkables(Exception):
    def __init__(self):
        super().__init__(f"No workables were passed to the proc")



# class FluidSines(FTISAnalyser):
#     def __init__(self,
#     bandwidth=76,
#     birthhighthreshold=-60,
#     birthlowthreshold=-24,
#     detectionthreshold=-96,
#     fftsettings=[1024, 512, 1024],
#     mintracklen=15,
#     trackingmethod=0,
#     trackfreqrange=50.0,
#     trackmagrange=15.0,
#     trackprob=0.5
#     ):
#         super().__init__()
#         self.bandwidth = bandwidth,
#         self.birthhighthreshold,
#         self.birthlowthreshold,
#         self.detectionthreshold,
#         self.fftsettings,
#         self.mintracklen

#     def analyse(self, workable):
#         out_folder = self.output / workable.name
#         out_folder.mkdir(exist_ok=True)

#         sines = out_folder / f"sines_{workable.name}"
#         residual = out_folder / f"residual_{workable.name}"

#         fluid.sines(
#             workable,
#             sines=sines,
#             residual=residual,
#             bandwidth=self.parameters["bandwidth"],
#             birthhighthreshold=self.parameters["birthhighthreshold"],
#             birthlowthreshold=self.parameters["birthlowthreshold"],
#             detectionthreshold=self.parameters["detectionthreshold"],
#             fftsettings=self.parameters["fftsettings"],
#             mintracklen=self.parameters["mintracklen"],
#             trackingmethod=self.parameters["trackmethod"],
#             trackfreqrange=self.parameters["trackfreqrange"],
#             trackmagrange=self.parameters["trackmagrange"],
#             trackprob=self.parameters["trackprob"],
#         )

#     def run(self):
#         workables = self.input
#         singleproc(self.name, self.analyse, workables)
#         self.output

# class FluidTransients(FTISAnalyser):
#     def __init__(self):
#         super().__init__(parent_process)
#         self.input_type = Ftypes.folder
#         self.output_type = Ftypes.folder

#     def analyse(self, workable):
#         out_folder = self.output / workable.name
#         out_folder.mkdir(exist_ok=True)

#         transients = out_folder / f"transients_{workable.name}"
#         residual = out_folder / f"residual_{workable.name}"

#         fluid.transients(
#             workable,
#             transients=transients,
#             residual=residual,
#             blocksize=self.parameters["blocksize"],
#             clumplength=self.parameters["clumplength"],
#             order=self.parameters["order"],
#             padsize=self.parameters["padsize"],
#             skew=self.parameters["skew"],
#             threshback=self.parameters["threshback"],
#             threshfwd=self.parameters["threshfwd"],
#             windowsize=self.parameters["windowsize"],
#         )

#     def run(self):
#         workables = self.input
#         multiproc(self.name, self.analyse, workables)
#         cleanup()



from ftis.common.analyser import FTISAnalyser
from ftis.common.io import write_json, read_json, peek, get_sr
from ftis.common.proc import singleproc
from ftis.common.conversion import samps2ms
from shutil import copyfile
from pathlib import Path
import soundfile as sf

class CollapseAudio(FTISAnalyser):
    def __init__(self):
        super().__init__()

    def collapse(self, workable):
        out = self.outfolder / Path(workable).name
        raw, sr = sf.read(workable, dtype="float32")
        audio = None
        if raw.ndim == 1:
            audio = raw
        else:
            audio = raw.transpose().sum(axis=0) / raw.ndim
        sf.write(out, audio, sr, "PCM_32")

    def run(self):
        self.outfolder = (
            self.process.sink / 
            f"{self.order}.{self.suborder}-{self.parent_string}"
        )
        self.outfolder.mkdir(exist_ok=True)
        singleproc(self.name, self.collapse, self.input)
        self.output = [x for x in self.outfolder.iterdir() if x.suffix == ".wav"]


class ExplodeAudio(FTISAnalyser):
    def __init__(self, cache=False):
        super().__init__(cache=cache)
        self.dump_type = ".json"

    def load_cache(self):
        d = read_json(self.dump_path)
        self.output = [Path(x) for x in d["corpus_items"]]

    def dump(self):
        d = {"corpus_items": [str(x) for x in self.output]}
        write_json(self.dump_path, d)

    def segment(self, workable):
        slices = [int(x) for x in self.input[str(workable)]]
        if len(slices) == 1:
            copyfile(workable, self.outfolder / f"{workable.stem}_0.wav")
        else:
            data, sr = sf.read(workable, dtype="float32")
            # Append the right boundary if it isnt already there
            if data.shape[0] != slices[-1]:
                slices.append(data.shape[0])

            for i, (start, end) in enumerate(zip(slices, slices[1:])):
                segment = data[start:end]
                sf.write(self.outfolder / f"{workable.stem}_{i}.wav", segment, sr, "PCM_32")

    def run(self):
        self.outfolder = (
            self.process.sink / 
            f"{self.order}.{self.suborder}-{self.parent_string}"
        )
        self.outfolder.mkdir(exist_ok=True)
        workables = [Path(x) for x in self.input.keys()]
        singleproc(self.name, self.segment, workables)
        self.output = [x for x in self.output_folder.iterdir() if x.suffix in [".wav", ".aiff", ".aif"]]



from setuptools import setup, find_packages
from pathlib import Path

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="ftis",
    version="2.0.0a",
    author="James Bradbury",
    url="https://github.com/jamesb93/ftis",
    license="GLPv3+",
    author_email="jamesbradbury93@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    description="The finding things in stuff package.",
    packages=find_packages(),
    install_requires=[
        "Soundfile",
        "hdbscan",
        "umap-learn",
        "sklearn",
        "python-flucoma",
        "scipy",
        "rich",
        "librosa",
        "jinja2",
    ],
)



import datetime
import logging
from pathlib import Path
from rich.console import Console
from rich.markdown import Markdown
from ftis.common.io import write_json, read_json
from ftis.common.exceptions import InvalidSource
from ftis.common.utils import ignored_keys, create_hash


class FTISProcess:
    """Class that represents the life cycle of a 'FTIS' execution"""

    def __init__(self, source=None, sink=None, mode="chain"):
        self.sink = Path(sink).expanduser().resolve()
        self.corpus = source
        self.source = source.items  # FIXME this is confusing
        self.chain = []
        self.logger = logging.getLogger(__name__)
        self.console = Console()
        self.mode = mode
        self.metadata = {}
        self.prev_meta = None

    def setup(self):
        self.sink.mkdir(exist_ok=True, parents=True)

        # Create a place to store microcached results
        self.cache = self.sink / ".cache"
        self.cache.mkdir(exist_ok=True)

        # Setup logging and meta path
        self.metapath = self.sink / "metadata.json"
        logfile_path = self.sink / "logfile.log"

        # Read in previous metadata if exists
        try:
            self.prev_meta = read_json(self.metapath)
        except FileNotFoundError:
            self.prev_meta = None

        self.logger.setLevel(logging.DEBUG)

        if logfile_path.exists():
            logfile_path.unlink()

        logfile_handler = logging.FileHandler(logfile_path)
        formatter = logging.Formatter("%(asctime)s : %(levelname)s : %(name)s : %(message)s")
        logfile_handler.setFormatter(formatter)
        self.logger.addHandler(logfile_handler)
        self.logger.debug("Logging initialised")

    def general_metadata(self):
        # Time
        self.metadata["time"] = datetime.datetime.now().strftime("%H:%M:%S | %B %d, %Y")

        # Analyser chain
        self.metadata["io"] = str([link.name for link in self.chain])

    def fprint(self, text):
        self.console.print(text, style="yellow underline")

    def add(self, *args):
        """Accepts any number of classes to chain together"""
        self.chain = args  # Lets store these classes somewhere

        analyser_params = {}
        for i, analyser in enumerate(self.chain):
            analyser.order = i
            name = analyser.__class__.__name__

            analyser_params[f"{i}_{name}"] = {
                k: v for k, v in vars(analyser).items() if k not in ignored_keys
            }
            analyser.process = self
            analyser.set_dump()

        self.metadata["analyser"] = analyser_params
        # self.order_hash = create_hash(*[x.name for x in self.chain]) #FIXME moot?

    def run_analysers(self):
        for i, analyser in enumerate(self.chain):
            if self.mode == "chain":
                if i == 0:
                    analyser.input = self.source
                else:
                    analyser.input = self.chain[i - 1].output
            else:
                analyser.input = self.source
            analyser.create_identity()
            self.metadata["analyser"][f"{i}_{analyser.name}"]["identity_hash"] = analyser.identity_hash
            analyser.do()

    def run(self):
        self.setup()
        md = "# **** FTIS v1.1.1 ****"
        md += f"\n\n**Source: {self.corpus.path}**"
        md += f"\n\n**Sink: {self.sink}**"
        md += "\n\n---------------------"
        md += "\n\nBeginning processing..."
        self.console.print(Markdown(md))
        print("\n")
        self.run_analysers()
        self.general_metadata()
        write_json(self.metapath, self.metadata)



from pathlib import Path
import numpy as np
from ftis.common.exceptions import NoCorpusSource, InvalidSource
from ftis.common.analyser import FTISAnalyser
from ftis.common.proc import singleproc
from ftis.common.io import write_json, read_json, get_duration
from ftis.common.utils import create_hash
from flucoma.utils import get_buffer
from flucoma.fluid import stats, loudness
from rich.progress import Progress
from typing import List


class Corpus:
    def __init__(self, path: str = "", file_type: List[str] = [".wav", ".aiff", ".aif"]):
        self.path = path
        self.name = self.__class__.__name__
        self.file_type = file_type
        self.items: List = []
        self.is_filtering: bool = False
        self.chain = {}
        self.get_items()

    def __add__(self, right):
        try:
            self.items += right.items  # this is the fastest way to merge in place
        except AttributeError:
            raise
        return self

    def __rshift__(self, right):
        self.chain[right] = None
        # right.order = 2
        return right


    def walk_chain(self) -> None:
        # Pass output to the input of all of connected things
        for forward_connection in self.chain:
            forward_connection.input = self.items
            forward_connection.walk_chain()


    def get_items(self) -> None:
        if self.path == "":
            raise NoCorpusSource("Please provide a valid path for the corpus")

        self.path = Path(self.path).expanduser().resolve()

        if not self.path.exists():
            raise InvalidSource(self.path)

        if self.path.is_dir():
            self.items = [x for x in self.path.iterdir() if x.suffix in self.file_type]
        else:
            self.items = [str(self.path)]

    def startswith(self, prefix: str):
        with Progress() as progress:
            task = progress.add_task("[cyan]Corpus Filtering: Name (startswith)", total=len(self.items))
            temp = []
            for x in self.items:
                if str(x.stem).startswith(prefix):
                    temp.append(x)
                progress.update(task, advance=1)
            self.items = temp
        return self

    def endswith(self, suffix: str):
        with Progress() as progress:
            task = progress.add_task("[cyan]Corpus Filtering: Name (endswith)", total=len(self.items))
            temp = []
            for x in self.items:
                if str(x.stem).endswith(suffix):
                    temp.append(x)
                progress.update(task, advance=1)
            self.items = temp
        return self

    def has(self, has: str):
        with Progress() as progress:
            task = progress.add_task("[cyan]Corpus Filtering: Name (has)", total=len(self.items))
            temp = []
            for x in self.items:
                if has in str(x):
                    temp.append(x)
                progress.update(task, advance=1)
            self.items = temp
        return self

    def loudness(self, min_loudness: int = 0, max_loudness: int = 100):
        hopsize = 4410
        windowsize = 17640
        with Progress() as progress:
            task = progress.add_task("[cyan]Corpus Filtering: Loudness", total=len(self.items))

            median_loudness = {}
            for x in self.items:
                hsh = create_hash(x, hopsize, windowsize)

                # Make sure a sane temporary path exists
                tmp = Path("/tmp") / "ftis_cache"
                tmp.mkdir(exist_ok=True)

                cache = tmp / f"{hsh}.npy"
                if not cache.exists():
                    med_loudness = get_buffer(
                        stats(loudness(x, hopsize=hopsize, windowsize=windowsize)), "numpy"
                    )
                    np.save(cache, med_loudness)
                else:
                    med_loudness = np.load(cache, allow_pickle=True)

                median_loudness[str(x)] = med_loudness[0][5]
                progress.update(task, advance=1)

            # Get percentiles and filter
            vals = np.array([x for x in median_loudness.values()])
            min_perc = np.percentile(vals, min_loudness)
            max_perc = np.percentile(vals, max_loudness)
            self.items = [k for k, v in median_loudness.items() if v <= max_perc and v >= min_perc]
        return self

    @staticmethod
    def filter_duration(x, low: float, high: float) -> bool:
        hsh = create_hash(x, low, high)
        tmp = Path("/tmp") / "ftis_cache"
        tmp.mkdir(exist_ok=True)

        cache = tmp / f"{hsh}.npy"
        if not cache.exists():
            dur = get_duration(x)
            np.save(cache, dur)
        else:
            dur = np.load(cache)
        return dur < high and dur > low

    def duration(self, min_duration: int = 0, max_duration: int = 36000):
        # TODO handle min/max types that can come in so you can do percentages
        self.is_filtering = True
        self.items = [x for x in self.items if self.filter_duration(x, min_duration, max_duration)]
        return self


class Analysis:
    # TODO This could be merged directly into the corpus class where it would
    # directly determine the type from the extension
    """This class lets you directly use analysis as an entry point to FTIS"""

    def __init__(self, path=""):
        self.path = path
        self.items = None
        self.get_items()

    def get_items(self):
        if self.path == "":
            raise NoCorpusSource("Please provide a valid path for the analysis")

        self.path = Path(self.path).expanduser().resolve()

        if not self.path.exists():
            raise InvalidSource(self.path)

        self.items = path



from ftis.common.analyser import FTISAnalyser
from ftis.common.io import write_json, read_json
from ftis.common.proc import singleproc
from multiprocessing import Manager
from scipy.signal import savgol_filter
from scipy.io import wavfile
from sklearn.cluster import AgglomerativeClustering
import numpy as np


class ClusteredNMF(FTISAnalyser):
    def __init__(
        self,
        iterations=100,
        components=10,
        fftsettings=[4096, 1024, 4096],
        smoothing=11,
        polynomial=2,
        min_cluster_size=2,
        min_samples=2,
        cluster_selection_method="eom",
        cache=False,
    ):
        super().__init__(cache=cache)
        self.components = components
        self.iterations = iterations
        self.fftsettings = fftsettings
        self.smoothing = smoothing
        self.polynomial = polynomial
        self.min_cluster_size = min_cluster_size
        self.min_samples = min_samples
        self.cluster_selection_method = cluster_selection_method
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def analyse(self, workable):
        nmf = fluid.nmf(
            workable,
            iterations=self.iterations,
            components=self.components,
            fftsettings=self.fftsettings,
        )
        bases = get_buffer(nmf.bases, "numpy")
        bases_smoothed = np.zeros_like(bases)

        for i, x in enumerate(bases):
            bases_smoothed[i] = savgol_filter(x, self.smoothing, self.polynomial)

        clusterer = hdbscan.HDBSCAN(
            min_cluster_size=self.min_cluster_size,
            min_samples=self.min_samples,
            cluster_selection_method=self.cluster_selection_method,
        )

        cluster_labels = clusterer.fit_predict(bases_smoothed)
        unique_clusters = list(dict.fromkeys(cluster_labels))

        sound = get_buffer(nmf.resynth, "numpy")

        for x in unique_clusters:
            summed = np.zeros_like(sound[0])  # make an empty numpy array of same size
            base = workable.name
            output = self.output / f"{base}_{x}.wav"
            for idx, cluster in enumerate(cluster_labels):
                if cluster == x:
                    summed += sound[idx]
            wavfile.write(output, 44100, summed)

    def run(self):
        self.output = self.process.sink / f"{self.order}_{self.__class__.__name__}"
        self.output.mkdir(exist_ok=True)
        workables = [
            k for k in self.input.iterdir() if k.name != ".DS_Store" and k.is_file() and k.suffix == ".wav"
        ]
        singleproc(self.name, self.analyse, workables)


class ClusteredSegmentation(FTISAnalyser):
    def __init__(self, numclusters=2, windowsize=4, cache=False):
        super().__init__(cache=cache)
        self.numclusters = numclusters
        self.windowsize = windowsize
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def analyse(self, workable):
        slices = self.input[workable]
        slices = [int(x) for x in slices]
        if len(slices) == 1:
            self.buffer[workable] = slices
            return
        count = 0
        standardise = StandardScaler()
        model = AgglomerativeClustering(n_clusters=self.numclusters)

        while (count + self.windowsize) <= len(slices):
            indices = slices[count : count + self.windowsize]  # create a section of the indices in question
            data = []
            for i, (start, end) in enumerate(zip(indices, indices[1:])):

                mfcc = mfcc(
                    workable,
                    fftsettings=[2048, -1, -1],
                    startframe=start,
                    numframes=end - start,
                )

                stats = get_buffer(stats(mfcc, numderivs=1), "numpy")

                data.append(stats.flatten())

            data = standardise.fit_transform(data)

            cluster = model.fit(data)
            cur = -2
            for j, c in enumerate(cluster.labels_):
                prev = cur
                cur = c
                if cur == prev:
                    try:
                        slices.pop(j + count)
                    except IndexError:
                        pass  # FIXME why are some indices erroring?
            count += 1
        self.buffer[workable] = slices

    def run(self):
        self.buffer = Manager().dict()
        workables = [x for x in self.input]
        singleproc(self.name, self.analyse, workables)
        self.output = dict(self.buffer)



from ftis.common.analyser import FTISAnalyser
from ftis.common.proc import staticproc
from ftis.common.io import write_json, read_json
from sklearn.neighbors import KDTree as SKKDTree
from sklearn.cluster import AgglomerativeClustering as AggCluster
from joblib import dump as jdump
import numpy as np


class AgglomerativeClustering(FTISAnalyser):
    def __init__(self, numclusters=3, cache=False):
        super().__init__(cache=cache)
        self.numclusters = numclusters
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def analyse(self):
        keys = [x for x in self.input.keys()]
        values = [x for x in self.input.values()]

        data = np.array(values)

        db = AggCluster(n_clusters=self.numclusters).fit(data)

        self.output = {}

        for audio, cluster in zip(keys, db.labels_):
            if str(cluster) in self.output:
                self.output[str(cluster)].append(audio)
            else:
                self.output[str(cluster)] = [audio]

    def run(self):
        staticproc(self.name, self.analyse)


class HDBSCAN(FTISAnalyser):
    def __init__(self, minclustersize=2, minsamples=1, cache=False):
        super().__init__(cache=cache)
        self.minclustersize = minclustersize
        self.minsamples = minsamples
        self.dump_type = ".json"

    def load_cache(self):
        self.output = read_json(self.dump_path)

    def dump(self):
        write_json(self.dump_path, self.output)

    def analyse(self):
        keys = [x for x in self.input.keys()]
        values = [x for x in self.input.values()]

        data = np.array(values)

        db = HdbscanClustering(
            min_cluster_size=self.minclustersize,
            min_samples=self.minsamples,
        ).fit(data)

        self.output = {}

        for audio, cluster in zip(keys, db.labels_):
            if str(cluster) in self.output:
                self.output[str(cluster)].append(audio)
            else:
                self.output[str(cluster)] = [audio]

    def run(self):
        staticproc(self.name, self.analyse)


class KDTree(FTISAnalyser):
    def __init__(self, cache=False):
        super().__init__(cache=cache)

    def dump(self):
        jdump(self.model, self.model_dump)

    def analyse(self):
        data = [v for v in self.input.values()]
        keys = [k for k in self.input.keys()]
        data = np.array(data)
        self.model = SKKDTree(data)

    def run(self):
        staticproc(self.name, self.analyse)
        self.output = self.model_dump
